from scipy.interpolate import interp1d
import pandas as pd
import scipy
import pickle as pkl
import numpy as np
import datetime

from CosRayModifiedISO import CosRayModifiedISO

print("WARNING: currently unknown whether the reported spectral weighting factor should be in terms of energy or rigidity")

class rigiditySpectrum():
    """
    Base class for rigidity spectra.
    """

    def __init__(self):
        """
        Initialize the rigidity spectrum.
        """
        pass

    def __call__(self, x: float) -> float:
        """
        Evaluate the rigidity spectrum at a given rigidity.

        Parameters:
        - x: float
            The rigidity.

        Returns:
        - float
            The value of the rigidity spectrum.
        """
        return self.rigiditySpec(x)
    
    def __add__(self, right: 'rigiditySpectrum') -> 'rigiditySpectrum':
        """
        Add two rigidity spectra.

        Parameters:
        - right: rigiditySpectrum
            The other rigidity spectrum.

        Returns:
        - rigiditySpectrum
            The sum of the two rigidity spectra.
        """
        summed_spectrum = rigiditySpectrum()
        summed_spectrum.rigiditySpec = lambda x: self.rigiditySpec(x) + right.rigiditySpec(x)
        return summed_spectrum

class powerLawSpectrum(rigiditySpectrum):
    """
    Power law rigidity spectrum.
    """

    rigiditySpec = lambda self, x: self.normalisationFactor * (x ** self.spectralIndex)

    def __init__(self, normalisationFactor: float, spectralIndex: float):
        """
        Initialize the power law spectrum.

        Parameters:
        - normalisationFactor: float
            The normalization factor.
        - spectralIndex: float
            The spectral index.
        """
        self.normalisationFactor = normalisationFactor
        self.spectralIndex = spectralIndex

class interpolatedInputFileSpectrum(rigiditySpectrum):
    """
    Interpolated rigidity spectrum from an input file.
    """

    def __init__(self, inputFileName: str):
        """
        Initialize the interpolated spectrum.

        Parameters:
        - inputFileName: str
            The path to the input file.
        """
        self.inputFilename = inputFileName
        self.rigiditySpec = self.readSpecFromCSV(self.inputFilename)

    def readSpecFromCSV(self, inputFileName: str) -> callable:
        """
        Read the spectrum from a CSV file.

        Parameters:
        - inputFileName: str
            The name of the input file.

        Returns:
        - callable
            The interpolated spectrum.
        """
        inputDF = pd.read_csv(inputFileName, header=None)
        rigidityList = inputDF[0]  # GV
        fluxList = inputDF[1]  # p/m2/sr/s/GV
        fluxListcm2 = fluxList / (100 ** 2)

        rigiditySpec = scipy.interpolate.interp1d(rigidityList, fluxListcm2, kind="linear",
                                                  fill_value=0.0, bounds_error=False)
        
        return rigiditySpec

class DLRmodelSpectrum(rigiditySpectrum):
    """
    DLR model rigidity spectrum.
    """

    def __init__(self, atomicNumber: int, date_and_time: 'datetime' = None, OULUcountRateInSeconds: float = None, W_parameter: float = None):
        """
        Initialize the DLR model spectrum.

        Parameters:
        - atomicNumber: int
            The atomic number of the particle.
        - date_and_time: datetime, optional
            The date and time for the spectrum.
        - OULUcountRateInSeconds: float, optional
            The OULU count rate in seconds.
        - W_parameter: float, optional
            The W parameter for the DLR model.
        """
        if not sum([(date_and_time is not None), (OULUcountRateInSeconds is not None), (W_parameter is not None)]) == 1:
            print("Error: exactly one supplied input out of the date and time, OULU count rate per second or the W parameter, to the DLR model spectrum must be given!")
            raise Exception

        if date_and_time is not None:
            self._generatedSpectrumDF = CosRayModifiedISO.getSpectrumUsingTimestamp(timestamp=date_and_time, atomicNumber=atomicNumber)

        if W_parameter is not None:
            self._generatedSpectrumDF = CosRayModifiedISO.getSpectrumUsingSolarModulation(solarModulationWparameter=W_parameter, atomicNumber=atomicNumber)

        if OULUcountRateInSeconds is not None:
            self._generatedSpectrumDF =CosRayModifiedISO.getSpectrumUsingOULUcountRate(OULUcountRatePerSecond=OULUcountRateInSeconds, atomicNumber=atomicNumber)

        self.rigiditySpec = interp1d(x=self._generatedSpectrumDF["Rigidity (GV/n)"],
                                     y=self._generatedSpectrumDF["d_Flux / d_R (cm-2 s-1 sr-1 (GV/n)-1)"],
                                     kind="linear",
                                     bounds_error=False,
                                     fill_value=(0.0, 0.0))

        with open("DLRsavedSpectrum.pkl", "wb") as DLRspecFile:
            pkl.dump(self, DLRspecFile)

class CommonModifiedPowerLawSpectrum(rigiditySpectrum):
    """
    Common modified power law rigidity spectrum.
    """

    specIndexModification = lambda self, P: self.deltaGamma * (P - 1)
    rigiditySpec = lambda self, P: self.J0 * self.step_function(P, self.lowerLimit, self.upperLimit) * (P ** (-(self.gamma + self.specIndexModification(P)))) / (100 ** 2)  # cm-2 s-1 sr-1 GV-1 : converted from m-2 to cm-2

    def __init__(self, J0: float, gamma: float, deltaGamma: float, lowerLimit: float = -np.inf, upperLimit: float = np.inf):
        """
        Initialize the common modified power law spectrum.

        Parameters:
        - J0: float
            The normalization factor.
        - gamma: float
            The spectral index.
        - deltaGamma: float
            The modification factor for the spectral index.
        - lowerLimit: float, optional
            The lower limit for the rigidity.
        - upperLimit: float, optional
            The upper limit for the rigidity.
        """
        self.lowerLimit = lowerLimit
        self.upperLimit = upperLimit

        self.J0 = J0  # m-2 s-1 sr-1 GV-1
        self.gamma = gamma
        self.deltaGamma = deltaGamma

        with open("CommonsavedSpectrum.pkl", "wb") as CommonspecFile:
            pkl.dump(self, CommonspecFile)

    def step_function(self, rigidity: float, lowerLimit: float, upperLimit: float) -> float:
        """
        Step function for the rigidity spectrum.

        Parameters:
        - rigidity: float
            The rigidity.
        - lowerLimit: float
            The lower limit for the rigidity.
        - upperLimit: float
            The upper limit for the rigidity.

        Returns:
        - float
            The value of the step function.
        """
        if (rigidity >= lowerLimit) and (rigidity <= upperLimit):
            return 1.0
        else:
            return 0.0

class CommonModifiedPowerLawSpectrumSplit(rigiditySpectrum):
    """
    Common modified power law rigidity spectrum with split spectral index modification.
    """

    specIndexModification_high = lambda self, P: self.deltaGamma * (P - 1)
    specIndexModification_low = lambda self, P: self.deltaGamma * (P)
    specIndexModification = lambda self, P: self.specIndexModification_high(P) if P > 1.0 else self.specIndexModification_low(P)
    rigiditySpec = lambda self, P: self.J0 * (P ** (-(self.gamma + self.specIndexModification(P)))) / (100 ** 2)  # cm-2 s-1 sr-1 GV-1 : converted from m-2 to cm-2

    def __init__(self, J0: float, gamma: float, deltaGamma: float):
        """
        Initialize the common modified power law spectrum with split spectral index modification.

        Parameters:
        - J0: float
            The normalization factor.
        - gamma: float
            The spectral index.
        - deltaGamma: float
            The modification factor for the spectral index.
        """
        self.J0 = J0  # m-2 s-1 sr-1 GV-1
        self.gamma = gamma
        self.deltaGamma = deltaGamma

        with open("CommonsavedSpectrum.pkl", "wb") as CommonspecFile:
            pkl.dump(self, CommonspecFile)



import numpy as np

from .pitchAngleDistribution import pitchAngleDistribution
from .rigiditySpectrum import rigiditySpectrum

def jacobian_function_to_use(pitch_angle_in_radians: float) -> float:
    try:
        output_Jacobian_factor = 1 / np.sin(2.0 * pitch_angle_in_radians)
    except ZeroDivisionError:
        output_Jacobian_factor = 0.0
    return output_Jacobian_factor

class momentaDistribution():
    """
    Class representing a momenta distribution.
    """

    def __init__(self, rigidity_spectrum: rigiditySpectrum, pitch_angle_distribution: pitchAngleDistribution):
        """
        Initialize the momenta distribution.

        Parameters:
        - rigidity_spectrum: rigiditySpectrum
            The rigidity spectrum.
        - pitch_angle_distribution: pitchAngleDistribution
            The pitch angle distribution.
        """
        self.setRigiditySpectrum(rigidity_spectrum)
        self.setPitchAngleDistribution(pitch_angle_distribution)

    # setter methods

    def setPitchAngleDistribution(self, pitch_angle_distribution: pitchAngleDistribution):
        """
        Set the pitch angle distribution.

        Parameters:
        - pitch_angle_distribution: pitchAngleDistribution
            The pitch angle distribution.
        """
        self.pitch_angle_distribution = pitch_angle_distribution

    def setRigiditySpectrum(self, rigidity_spectrum: rigiditySpectrum):
        """
        Set the rigidity spectrum.

        Parameters:
        - rigidity_spectrum: rigiditySpectrum
            The rigidity spectrum.
        """
        self.rigidity_spectrum = rigidity_spectrum

    # getter methods

    def getPitchAngleDistribution(self) -> pitchAngleDistribution:
        """
        Get the pitch angle distribution.

        Returns:
        - pitchAngleDistribution
            The pitch angle distribution.
        """
        return self.pitch_angle_distribution

    def getRigiditySpectrum(self) -> rigiditySpectrum:
        """
        Get the rigidity spectrum.

        Returns:
        - rigiditySpectrum
            The rigidity spectrum.
        """
        return self.rigidity_spectrum

    # other Methods

    def __call__(self, pitchAngle: float, rigidity: float) -> float:
        """
        Evaluate the momenta distribution.

        Parameters:
        - pitchAngle: float
            The pitch angle.
        - rigidity: float
            The rigidity.

        Returns:
        - float
            The value of the momenta distribution.
        """
        pitch_angle_weighting_factor = self.pitch_angle_distribution(pitchAngle, rigidity) #* jacobian_function_to_use(pitchAngle)
        rigidity_weighting_factor = self.rigidity_spectrum(rigidity)

        full_rigidity_pitch_weighting_factor = pitch_angle_weighting_factor * rigidity_weighting_factor

        return full_rigidity_pitch_weighting_factor


#import cudf.pandas
#cudf.pandas.install()

import numpy as np
import pandas as pd
import datetime as dt
import sys
import ParticleRigidityCalculationTools as PRCT
from joblib import Memory
import tqdm
tqdm.tqdm.pandas()
from pandarallel import pandarallel
pandarallel.initialize(progress_bar=True)
from spacepy.coordinates import Coords as spaceCoords
from spacepy.time import Ticktock as spaceTicktock
import numba

from .spectralCalculations.particleDistribution import particleDistribution
from .spectralCalculations.momentaDistribution import momentaDistribution

memory_asymp_dirs = Memory("./cacheAsymptoticDirectionOutputs", verbose=0)

m0 = 1.67262192e-27 #kg
c = 299792458.0 #m/s
protonCharge = 1.60217663e-19 #C
protonRestEnergy = m0 * (c**2)

global get_apply_method
def get_apply_method(DF_or_Series: pd.DataFrame) -> callable:
    """
    Determine the appropriate apply method based on the debugging state.

    Parameters:
    - DF_or_Series: pd.DataFrame
        The DataFrame or Series to apply the method to.

    Returns:
    - callable
        The appropriate apply method.
    """
    if hasattr(sys, 'gettrace') and sys.gettrace() is not None:
        print("debug mode being used: setting AniMAIRE to use progress_apply rather than running in parallel!")
        apply_method = DF_or_Series.progress_apply
    else:
        #print("not in debug mode: setting AniMAIRE to use parallel_apply!")
        apply_method = DF_or_Series.parallel_apply
        
    return apply_method

def generate_asymp_dir_DF(dataframeToFillFrom: pd.DataFrame, IMFlatitude: float, IMFlongitude: float, datetime_to_run_across_UTC: dt.datetime, cache: bool) -> pd.DataFrame:
    """
    Generate a DataFrame of asymptotic directions with pitch angles.

    Parameters:
    - dataframeToFillFrom: pd.DataFrame
        The DataFrame to fill from.
    - IMFlatitude: float
        The latitude of the Interplanetary Magnetic Field (IMF).
    - IMFlongitude: float
        The longitude of the Interplanetary Magnetic Field (IMF).
    - datetime_to_run_across_UTC: dt.datetime
        The datetime to run across in UTC.
    - cache: bool
        Whether to cache the results.

    Returns:
    - pd.DataFrame
        The DataFrame with asymptotic directions and pitch angles.
    """
    new_asymp_dir_DF = dataframeToFillFrom.copy()

    #new_asymp_dir_DF["Energy"] = PRCT.convertParticleRigidityToEnergy(dataframeToFillFrom["Rigidity"], particleMassAU = 1, particleChargeAU = 1)

    print("assigning asymptotic coordinates")
    if cache == False:
        asymptoticDirectionList = convertAsymptoticDirectionsToPitchAngle(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC)
    else:
        cachedConvertAsympDirFunc = memory_asymp_dirs.cache(convertAsymptoticDirectionsToPitchAngle)
        asymptoticDirectionList = cachedConvertAsympDirFunc(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC)

    print("successfully converted asymptotic directions")

    new_asymp_dir_DF["angleBetweenIMFinRadians"] = asymptoticDirectionList

    return new_asymp_dir_DF

def convertAsymptoticDirectionsToPitchAngle(dataframeToFillFrom: pd.DataFrame, IMFlatitude: float, IMFlongitude: float, datetime_to_run_across_UTC: dt.datetime) -> pd.Series:
    """
    Convert asymptotic directions to pitch angles.

    Parameters:
    - dataframeToFillFrom: pd.DataFrame
        The DataFrame to fill from.
    - IMFlatitude: float
        The latitude of the Interplanetary Magnetic Field (IMF).
    - IMFlongitude: float
        The longitude of the Interplanetary Magnetic Field (IMF).
    - datetime_to_run_across_UTC: dt.datetime
        The datetime to run across in UTC.

    Returns:
    - pd.Series
        The Series with pitch angles.
    """
    print("acquiring pitch angles...")
    pitch_angle_list = get_apply_method(dataframeToFillFrom)(lambda dataframe_row: get_pitch_angle_for_DF_analytic(IMFlatitude, IMFlongitude, dataframe_row["Lat"], dataframe_row["Long"]), axis=1)

    return pitch_angle_list

@numba.jit(nopython=True)
def get_pitch_angle_for_DF_analytic(IMFlatitude: float, IMFlongitude: float, asymptotic_dir_latitude: float, asymptotic_dir_longitude: float) -> float:
    """
    Calculate the pitch angle for a DataFrame row analytically.

    Parameters:
    - IMFlatitude: float
        The latitude of the Interplanetary Magnetic Field (IMF).
    - IMFlongitude: float
        The longitude of the Interplanetary Magnetic Field (IMF).
    - asymptotic_dir_latitude: float
        The latitude of the asymptotic direction.
    - asymptotic_dir_longitude: float
        The longitude of the asymptotic direction.

    Returns:
    - float
        The pitch angle in radians.
    """
    IMFlatitude_rad = IMFlatitude * (np.pi / 180.0)
    IMFlongitude_rad = IMFlongitude * (np.pi / 180.0)
    asymptotic_dir_latitude_rad = asymptotic_dir_latitude * (np.pi / 180.0)
    asymptotic_dir_longitude_rad = asymptotic_dir_longitude * (np.pi / 180.0)

    cos_pitch_angle = (np.sin(asymptotic_dir_latitude_rad) * np.sin(IMFlatitude_rad)) + \
                      (np.cos(asymptotic_dir_latitude_rad) * np.cos(IMFlatitude_rad) * np.cos(asymptotic_dir_longitude_rad - IMFlongitude_rad))
    
    pitch_angle = np.arccos(cos_pitch_angle)

    return pitch_angle

def acquireWeightingFactors(asymptotic_direction_DF: pd.DataFrame, particle_dist: particleDistribution) -> pd.DataFrame:
    """
    Acquire weighting factors for asymptotic directions.

    Parameters:
    - asymptotic_direction_DF: pd.DataFrame
        The DataFrame with asymptotic directions.
    - particle_dist: particleDistribution
        The particle distribution.

    Returns:
    - pd.DataFrame
        The DataFrame with weighting factors.
    """
    momentaDist = particle_dist.momentum_distribution
    new_asymptotic_direction_DF = asymptotic_direction_DF.copy()

    # find weighting factors from the angles and rigidities
    #jacobian_function_to_use = lambda pitch_angle_in_radians:1/np.sin(2.0 * pitch_angle_in_radians)
        
    pitchAngleFunctionToUse = lambda row: momentaDist.getPitchAngleDistribution()(row["angleBetweenIMFinRadians"], row["Rigidity"])
    fullRigidityPitchWeightingFactorFunctionToUse = lambda row: momentaDist(row["angleBetweenIMFinRadians"], row["Rigidity"])

    print("calculating pitch angle weighting factors...")
    new_asymptotic_direction_DF["PitchAngleWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF)(pitchAngleFunctionToUse, axis=1)

    new_asymptotic_direction_DF["Filter"] = (new_asymptotic_direction_DF["Filter"] == 1) * 1

    print("calculating rigidity weighting factors...")
    new_asymptotic_direction_DF["RigidityWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF["Rigidity"])(momentaDist.getRigiditySpectrum())

    print("calculating rigidity + pitch combined weighting factors...")
    all_weighted_asymp_dirs = get_apply_method(new_asymptotic_direction_DF)(fullRigidityPitchWeightingFactorFunctionToUse, axis=1)
    new_asymptotic_direction_DF["fullRigidityPitchWeightingFactor"] = all_weighted_asymp_dirs * (new_asymptotic_direction_DF["Filter"] == 1)
    
    print("calculating energy + pitch combined weighting factors...")
    new_asymptotic_direction_DF["Energy"] = PRCT.convertParticleRigidityToEnergy(new_asymptotic_direction_DF["Rigidity"], 
                                                                      particleMassAU=particle_dist.particle_species.atomicMass, 
                                                                      particleChargeAU=particle_dist.particle_species.atomicNumber)
    energySpectrum = PRCT.convertParticleRigiditySpecToEnergySpec(new_asymptotic_direction_DF["Rigidity"],
                                                                  new_asymptotic_direction_DF["fullRigidityPitchWeightingFactor"],
                                                                  particleMassAU=particle_dist.particle_species.atomicMass, 
                                                                  particleChargeAU=particle_dist.particle_species.atomicNumber)

    new_asymptotic_direction_DF["fullEnergyPitchWeightingFactor"] = energySpectrum["Energy distribution values"]

    return new_asymptotic_direction_DF

def calculatePitchAngle_from_IMF_dir(interplanetary_mag_field: spaceCoords, asymptotic_direction: spaceCoords, datetime_to_run_across_UTC: dt.datetime) -> float:
    """
    Calculate the pitch angle from the Interplanetary Magnetic Field (IMF) direction.

    Parameters:
    - interplanetary_mag_field: spaceCoords
        The coordinates of the interplanetary magnetic field.
    - asymptotic_direction: spaceCoords
        The coordinates of the asymptotic direction.
    - datetime_to_run_across_UTC: dt.datetime
        The datetime to run across in UTC.

    Returns:
    - float
        The pitch angle in radians.
    """
    cartesianAsympDir = asymptotic_direction.convert("GEO", "car").data[0]
    GEOIMF = interplanetary_mag_field
    GEOIMF.ticks = spaceTicktock(datetime_to_run_across_UTC, "UTC")
    cartesianIMF = GEOIMF.convert("GEO", "car").data[0]
    return calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF)

def calculatePitchAngle(momentaDist: momentaDistribution, dfRow: pd.Series, datetime_to_run_across_UTC: dt.datetime) -> float:
    """
    Calculate the pitch angle.

    Parameters:
    - momentaDist: momentaDistribution
        The momenta distribution.
    - dfRow: pd.Series
        The DataFrame row.
    - datetime_to_run_across_UTC: dt.datetime
        The datetime to run across in UTC.

    Returns:
    - float
        The pitch angle in radians.
    """
    cartesianAsympDir = dfRow["Asymptotic Direction"].convert("GEO", "car").data[0]
    GEOIMF = momentaDist.getPitchAngleDistribution().interplanetary_mag_field
    GEOIMF.ticks = spaceTicktock(datetime_to_run_across_UTC, "UTC")
    cartesianIMF = GEOIMF.convert("GEO", "car").data[0]
    return calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF)

def calculateAngleBetweenTheSpaceVectors(cartesianAsympDir: np.ndarray, cartesianIMF: np.ndarray) -> float:
    """
    Calculate the angle between two space vectors.

    Parameters:
    - cartesianAsympDir: np.ndarray
        The Cartesian coordinates of the asymptotic direction.
    - cartesianIMF: np.ndarray
        The Cartesian coordinates of the interplanetary magnetic field.

    Returns:
    - float
        The angle between the vectors in radians.
    """
    dotProduct = np.dot(cartesianAsympDir, cartesianIMF)
    amplitude = np.linalg.norm(cartesianAsympDir) * np.linalg.norm(cartesianIMF)
    angleBetweenIMF = np.arccos(dotProduct / amplitude)
    return angleBetweenIMF



import numpy as np
from spacepy.coordinates import Coords as spaceCoords
import pickle as pkl
import copy

class pitchAngleDistribution():
    """
    Base class for pitch angle distributions.
    """

    def __init__(self, 
                 pitch_angle_distribution: callable, 
                 reference_latitude_in_GSM: float, 
                 reference_longitude_in_GSM: float):
        """
        Initialize the pitch angle distribution.

        Parameters:
        - pitch_angle_distribution: callable
            Function describing the pitch angle distribution.
        - reference_latitude_in_GSM: float
            Reference latitude in GSM coordinates.
        - reference_longitude_in_GSM: float
            Reference longitude in GSM coordinates.
        """
        self.pitchAngleDistFunction = pitch_angle_distribution

        self.interplanetary_mag_field = spaceCoords([100.0,
                                                     reference_latitude_in_GSM, 
                                                     reference_longitude_in_GSM],
                                                     "GSM","sph")

    def __call__(self, pitchAngle: float, rigidity: float) -> float:
        """
        Evaluate the pitch angle distribution at a given pitch angle and rigidity.

        Parameters:
        - pitchAngle: float
            The pitch angle.
        - rigidity: float
            The rigidity.

        Returns:
        - float
            The value of the pitch angle distribution.
        """
        return self.pitchAngleDistFunction(pitchAngle, rigidity)
    
    def __add__(self, right: 'pitchAngleDistribution') -> 'pitchAngleDistribution':
        """
        Add two pitch angle distributions.

        Parameters:
        - right: pitchAngleDistribution
            The other pitch angle distribution.

        Returns:
        - pitchAngleDistribution
            The sum of the two pitch angle distributions.
        """
        summed_dist = copy.deepcopy(self)
        summed_dist.pitchAngleDistFunction = lambda x, y: self.pitchAngleDistFunction(x, y) + right.pitchAngleDistFunction(x, y)
        return summed_dist
    
    def __mul__(self, right: float) -> 'pitchAngleDistribution':
        """
        Multiply the pitch angle distribution by a scalar.

        Parameters:
        - right: float
            The scalar.

        Returns:
        - pitchAngleDistribution
            The scaled pitch angle distribution.
        """
        multiplied_dist = copy.deepcopy(self)
        multiplied_dist.pitchAngleDistFunction = lambda x, y: right * self.pitchAngleDistFunction(x, y)
        return multiplied_dist

    __rmul__ = __mul__

class cosinePitchAngleDistribution(pitchAngleDistribution):
    """
    Cosine pitch angle distribution.
    """

    pitchAngleDistFunction = lambda self, pitchAngle, rigidity: np.abs(0.5 * np.sin(2 * pitchAngle))

    def __init__(self):
        """
        Initialize the cosine pitch angle distribution.
        """
        pass

class isotropicPitchAngleDistribution(pitchAngleDistribution):
    """
    Isotropic pitch angle distribution.
    """

    pitchAngleDistFunction = lambda self, pitchAngle, rigidity: 1

    def __init__(self):
        """
        Initialize the isotropic pitch angle distribution.
        """
        pass

class gaussianPitchAngleDistribution(pitchAngleDistribution):
    """
    Gaussian pitch angle distribution.
    """

    pitchAngleDistFunction = lambda self, pitchAngle, rigidity: self.normFactor * np.exp(-(pitchAngle - self.alpha)**2 / (self.sigma**2))

    def __init__(self, normFactor: float, sigma: float, alpha: float = 0.0):
        """
        Initialize the Gaussian pitch angle distribution.

        Parameters:
        - normFactor: float
            The normalization factor.
        - sigma: float
            The standard deviation of the Gaussian distribution.
        - alpha: float, optional
            The mean of the Gaussian distribution.
        """
        self.normFactor = normFactor
        self.sigma = sigma
        self.alpha = alpha
        with open("CommonsavedPAD.pkl", "wb") as CommonPADFile:
            pkl.dump(self, CommonPADFile)

class gaussianBeeckPitchAngleDistribution(pitchAngleDistribution):
    """
    Gaussian Beeck pitch angle distribution.
    """

    pitchAngleDistFunction = lambda self, pitchAngle_radians, rigidity: self.normFactor * np.exp((-0.5 * (pitchAngle_radians - (np.sin(pitchAngle_radians) * np.cos(pitchAngle_radians)))) / \
                                                                (self.A - (0.5 * (self.A - self.B) * (1 - np.cos(pitchAngle_radians)))))

    def __init__(self, normFactor: float, A: float, B: float):
        """
        Initialize the Gaussian Beeck pitch angle distribution.

        Parameters:
        - normFactor: float
            The normalization factor.
        - A: float
            Parameter A for the distribution.
        - B: float
            Parameter B for the distribution.
        """
        self.normFactor = normFactor
        self.A = A
        self.B = B


import numpy as np
import pandas as pd
from pandarallel import pandarallel
from atmosphericRadiationDoseAndFlux import doseAndFluxCalculator as DAFcalc
from scipy.interpolate import interp1d
from scipy.integrate import trapezoid
import ParticleRigidityCalculationTools as PRCT

from .data.NM64_responses import get_NM64_response_value_altitude, response_value_epns
from .AsymptoticDirectionProcessing import acquireWeightingFactors, get_apply_method
from .spectralCalculations.particleDistribution import particleDistribution
import dask.dataframe as dd

pandarallel.initialize(progress_bar=True)

class singleParticleEngineInstance:
    """
    Class to handle single particle engine instance for dose rate calculations.
    """

    def __init__(self, 
                 particle_distribution: particleDistribution, 
                 dfofAsymptoticDirections: pd.DataFrame, 
                 list_of_altitudes_in_km: list[float],
                 generate_NM_count_rates: bool):
        """
        Initialize the instance with necessary parameters.
        """
        self.particle_distribution = particle_distribution
        self.list_of_altitudes_in_km = list_of_altitudes_in_km
        self.dfOfAllAsymptoticDirections = dfofAsymptoticDirections
        self.generate_NM_count_rates = generate_NM_count_rates
        self.rigiditySpectrumParamDict = {}
        self.pitchAngleDistributionParamDict = {}

    def getThePitchAngleDistribution(self) -> object:
        """
        Get the pitch angle distribution.
        """
        print("Acquiring pitch angle distributions...")
        pitchAngleDistToUse = self.particle_distribution.momentum_distribution.pitch_angle_distribution
        try:
            print("IMF latitude set to", self.IMFlatitude)
            print("IMF longitude set to", self.IMFlongitude)
        except AttributeError:
            self.IMFlatitude = -1000.0 # distribution is probably isotropic: while a better solution is probably necessary, set to dummy variables for now.
            self.IMFlongitude = -1000.0
        pitchAngleDistToUse.setInterplanetaryMagFieldDirection(self.IMFlatitude, self.IMFlongitude)
        return pitchAngleDistToUse

    def getAsymptoticDirsAndRun(self, use_default_9_zeniths_azimuths: bool = False, record_full_output: bool = False, **kwargs) -> pd.DataFrame:
        """
        Acquire asymptotic directions and run calculations.
        Parameters:
          use_default_9_zeniths_azimuths (bool): flag for using default directions (unused here).
          record_full_output (bool): if True, attach additional output, including asymptotic directions as well as the calculated weighting factors for each asymptotic direction in attrs.
          **kwargs: additional keyword arguments.
        """
        self.acquireDFofAllAsymptoticDirections()
        sortedOutputDoseRates = self.runOverSpecifiedAltitudes(record_full_output=record_full_output)
        return sortedOutputDoseRates

    def runOverSpecifiedAltitudes(self, record_full_output: bool = False, **kwargs) -> pd.DataFrame:
        """
        Run calculations over specified altitudes.
        """
        print("Assigning pitch angle weighting factors...")
        df_with_weighting_factors_full_angles = acquireWeightingFactors(self.dfOfAllAsymptoticDirections, self.particle_distribution)
        df_with_weighting_factors_full_angles.to_pickle("df_with_weighting_factors_full_angles.pkl")

        df_with_weighting_factors = get_mean_weighting_factors_for_multi_angle_magcos_runs(df_with_weighting_factors_full_angles)

        print("Converting spectra and asymptotic directions to particle fluxes and dose rates...")
        sortedOutputDoseRates = self.calc_output_dose_flux(df_with_weighting_factors, self.list_of_altitudes_in_km, self.particle_distribution.particle_species.particleName)
        
        df_with_weighting_factors.to_pickle("weighting_factor_DF.pkl")
        if record_full_output:  
            sortedOutputDoseRates.attrs['weighting_factor_input_DF'] = dd.from_pandas(df_with_weighting_factors)
        
        print("Output dose rates calculated successfully!")
        return sortedOutputDoseRates

    def calc_output_dose_flux(self, asymp_dir_DF_with_weighting_factors: pd.DataFrame, list_of_altitudes_in_km: list[float], particle_name: str) -> pd.DataFrame:
        """
        Calculate output dose and flux.
        """
        spectrum_to_function_conversion_function = lambda row: interp1d(x=np.array(row["Rigidity"]),
                                                                        y=np.array(row["fullRigidityPitchWeightingFactor"]),
                                                                        bounds_error=False,
                                                                        fill_value=0.0)

        DFofSpectraForEachCoord = asymp_dir_DF_with_weighting_factors.groupby(["initialLatitude","initialLongitude"]).apply(spectrum_to_function_conversion_function)
        outputDoseRatesForAltitudeRange = get_apply_method(DFofSpectraForEachCoord)(lambda spectrum: DAFcalc.calculate_from_rigidity_spec(
                                                                                                inputRigidityDistributionFunctionGV=lambda x: float(spectrum(x)),
                                                                                                altitudesInkm=list_of_altitudes_in_km,
                                                                                                particleName=particle_name))

        outputDoseRatesOnlyDF = pd.concat(outputDoseRatesForAltitudeRange.tolist(), ignore_index=True)

        if self.generate_NM_count_rates:
            print("Calculating neutron monitor count rates...")
            outputDoseRatesOnlyDF["NM64_cr_unnorm"] = self.get_neutron_monitor_count_rates(list_of_altitudes_in_km, particle_name, DFofSpectraForEachCoord)
            print("Neutron monitor count rates successfully determined.")

        outputDoseRatesOnlyDF_with_lats_and_longs = self.reinsert_original_lats_and_longs(outputDoseRatesOnlyDF, outputDoseRatesForAltitudeRange)
        sortedOutputDoseRates = outputDoseRatesOnlyDF_with_lats_and_longs.sort_values(["latitude", "longitude", "altitude (km)"], ignore_index=True)
        return sortedOutputDoseRates

    def get_neutron_monitor_count_rates(self, list_of_altitudes_in_km: list[float], particle_name: str, DFofSpectraForEachCoord: pd.Series) -> np.ndarray:
        """
        Get neutron monitor count rates.
        """
        NM64_values_list = []

        for altitude_in_km in list_of_altitudes_in_km:
            NM64_vals_DF = get_apply_method(DFofSpectraForEachCoord)(lambda x: self.calculate_unnormed_NM64_cr(x, particle_name=particle_name, altitude_in_km=altitude_in_km)).values
            NM64_values_list.append(NM64_vals_DF)

        new_cr_unnorm_list = np.array(NM64_values_list).T.flatten()
        return new_cr_unnorm_list

    def calculate_unnormed_NM64_cr(self, interpolated_spectra_row: interp1d, particle_name: str, altitude_in_km: float = 0.0) -> float:
        """
        Calculate unnormalized NM64 count rates.
        """
        response_values = get_NM64_response_value_altitude(particle_name, response_value_epns, altitude_in_km)
        NM64_response_tabulated_2020_functional_epn = pd.DataFrame({"Energy_per_nucleon_GeV_per_n": response_value_epns,
                                                                    "Yield_per_m2_per_sr": response_values})

        if particle_name == "proton":
            NM64_response_tabulated_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(
                NM64_response_tabulated_2020_functional_epn["Energy_per_nucleon_GeV_per_n"] * 1000,
                fluxInEnergyMeVform=NM64_response_tabulated_2020_functional_epn["Yield_per_m2_per_sr"],
                particleMassAU=1,
                particleChargeAU=1)
        elif particle_name == "alpha":
            NM64_response_tabulated_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(
                NM64_response_tabulated_2020_functional_epn["Energy_per_nucleon_GeV_per_n"] * 1000 * 4,
                fluxInEnergyMeVform=NM64_response_tabulated_2020_functional_epn["Yield_per_m2_per_sr"],
                particleMassAU=4,
                particleChargeAU=2)
        else:
            raise Exception("ERROR: particle name did not match either proton or alpha!")

        NM64_response_to_use = NM64_response_tabulated_2020_functional
        spectra_weighting_factors_across_NM = NM64_response_to_use["Rigidity"].apply(interpolated_spectra_row)
        full_NM64_spectra_weighting_factor = spectra_weighting_factors_across_NM * NM64_response_to_use["Rigidity distribution values"]
        output_unnormalised_NM64_cr = trapezoid(full_NM64_spectra_weighting_factor, NM64_response_to_use["Rigidity"])
        return output_unnormalised_NM64_cr

    def reinsert_original_lats_and_longs(self, new_doserates_DF: pd.DataFrame, original_data_frame: pd.Series) -> pd.DataFrame:
        """
        Reinsert original latitudes and longitudes into the dataframe.
        """
        original_lats = original_data_frame.reset_index(level=[0, 1])["initialLatitude"]
        original_longs = original_data_frame.reset_index(level=[0, 1])["initialLongitude"]

        list_of_altitude_dfs = []
        for altitude_in_km in new_doserates_DF["altitude (km)"].unique():
            df_for_specific_altitude = new_doserates_DF[new_doserates_DF["altitude (km)"] == altitude_in_km].reset_index(drop=True).copy()
            df_for_specific_altitude["latitude"] = original_lats
            df_for_specific_altitude["longitude"] = original_longs
            list_of_altitude_dfs.append(df_for_specific_altitude)

        combined_df = pd.concat(list_of_altitude_dfs, ignore_index=True)
        # Reorder columns to place latitude and longitude on the left-hand side
        columns_order = ["latitude", "longitude"] + [col for col in combined_df.columns if col not in ["latitude", "longitude"]]
        return combined_df[columns_order]

def get_max_weighting_factors_for_multi_angle_magcos_runs(multi_angle_DF: pd.DataFrame) -> pd.DataFrame:
    """
    Get maximum weighting factors for multi-angle Magnetocosmics runs.
    """
    max_weighting_factor_DF = multi_angle_DF[multi_angle_DF[["Rigidity", "initialLatitude", "initialLongitude", "fullRigidityPitchWeightingFactor"]].groupby(
        ["Rigidity", "initialLatitude", "initialLongitude"], group_keys=False).apply(lambda x: x == x.max())["fullRigidityPitchWeightingFactor"]]
    return max_weighting_factor_DF.reset_index(drop=True)

def get_mean_weighting_factors_for_multi_angle_magcos_runs(multi_angle_DF: pd.DataFrame) -> pd.DataFrame:
    """
    Get mean weighting factors for multi-angle Magnetocosmics runs.
    """
    mean_weighting_factor_DF = multi_angle_DF.groupby(["Rigidity", "initialLatitude", "initialLongitude"], group_keys=False).mean()
    return mean_weighting_factor_DF.reset_index()